@@ -71,11 +71,13 @@
 
 #include "dbus/upstart.h"
 #include "com.ubuntu.Upstart.h"
+#include "control.h"
 
 
 #define BUILTIN_FSTAB   "/lib/init/fstab"
 
 #define BOREDOM_TIMEOUT 3
+#define ROOTDELAY 30
 
 enum exit {
 	EXIT_OK = 0,		/* Ok */
@@ -93,6 +95,7 @@ typedef enum {
 	TAG_SWAP,
 	TAG_NOWAIT,
 	TAG_SKIPPED,
+	TAG_TIMEOUT,
 } Tag;
 
 typedef enum {
@@ -227,6 +230,7 @@ void   plymouth_answer       (void *user_data, const char *keys,
 			      ply_boot_client_t *client);
 
 void   usr1_handler          (void *data, NihSignal *signal);
+int    set_dev_wait_time        (NihOption *option, const char *arg);
 
 
 /**
@@ -249,6 +253,8 @@ size_t num_virtual = 0;
 size_t num_virtual_mounted = 0;
 size_t num_swap = 0;
 size_t num_swap_mounted = 0;
+size_t num_timeout = 0;
+size_t num_timeout_mounted = 0;
 
 /**
  * newly_mounted:
@@ -355,6 +361,14 @@ static const char *plymouth_keys = NULL;
  **/
 static NihTimer *boredom_timer = NULL;
 
+/**
+ * device_ready_timer:
+ *
+ * Having waited over ROOTDELAY seconds for certain devices to get ready,
+ * exit mountall and trigger the upstart jobs for bringing up the device
+ **/
+
+NihTimer *device_ready_timer = NULL;
 
 /**
  * daemonise:
@@ -385,6 +399,16 @@ static int fsck_fix = FALSE;
  **/
 static int no_events = FALSE;
 
+/**
+ ** dev-wait-time
+ **
+ ** Set to ROOTDELAY if not set. 
+ ** Legal values are between 1sec to 2147483647 seconds.
+ **/
+static int dev_wait_time = FALSE;
+
+
+
 
 static void
 dequote (char *str)
@@ -669,6 +693,15 @@ parse_fstab (const char *filename)
 		Mount *         mnt;
 		nih_local char *fsname = NULL;
 
+		/* These are always symlinks into /run now.  Ignore any
+		 * explicit attempts to mount them because it vastly
+		 * complicates matters when trying to ensure that /run and
+		 * /var/run point to the same place.
+		 */
+		if (! strcmp (mntent->mnt_dir, "/var/run") ||
+		    ! strcmp (mntent->mnt_dir, "/var/lock"))
+			continue;
+
 		mnt = find_mount (mntent->mnt_dir);
 		if (mnt
 		    && strcmp (mntent->mnt_type, "swap")) {
@@ -812,6 +845,9 @@ parse_mountinfo_file (int reparsed)
 		if (! super_opts)
 			continue;
 
+		/* ignore ureadahead's mounting of debugfs */
+		if (! strcmp (mountpoint, "/var/lib/ureadahead/debugfs"))
+			continue;
 
 		opts = NIH_MUST (nih_sprintf (mounts, "%s,%s",
 					      mount_opts, super_opts));
@@ -979,7 +1015,9 @@ is_remote (Mount *mnt)
 	    || (! strcmp (mnt->type, "ncp"))
 	    || (! strcmp (mnt->type, "ncpfs"))
 	    || (! strcmp (mnt->type, "ocfs2"))
-	    || (! strcmp (mnt->type, "gfs"))) {
+	    || (! strcmp (mnt->type, "gfs"))
+	    || (! strcmp (mnt->type, "gfs2"))
+	    || (! strcmp (mnt->type, "ceph"))) {
 		return TRUE;
 	} else {
 		return FALSE;
@@ -1176,6 +1214,11 @@ mount_policy (void)
 			nih_info (_("%s is %s"), MOUNT_NAME (mnt), "local");
 			num_local++;
 			break;
+		case TAG_TIMEOUT:
+			nih_info (_("%s is %s"), MOUNT_NAME (mnt), "local with timeout");
+			num_local++;
+			num_timeout++;
+			break;
 		case TAG_REMOTE:
 			nih_info (_("%s is %s"), MOUNT_NAME (mnt), "remote");
 			num_remote++;
@@ -1229,7 +1272,13 @@ tag_mount (Mount *mnt,
 			}
 		} else {
 			if (! has_option (mnt, "nobootwait", FALSE)) {
-				tag = TAG_LOCAL;
+				if ( has_option (mnt, "timeout", FALSE))
+				{
+					tag = TAG_TIMEOUT; 
+				}
+				else
+					tag = TAG_LOCAL;
+	
 			} else {
 				tag = TAG_NOWAIT;
 			}
@@ -1252,6 +1301,9 @@ tag_mount (Mount *mnt,
 			if ((dep->tag == TAG_LOCAL)
 			    && (mnt->tag != TAG_REMOTE))
 				mnt->tag = TAG_LOCAL;
+			if ((dep->tag == TAG_TIMEOUT)
+			    && (mnt->tag != TAG_REMOTE))
+				mnt->tag = TAG_LOCAL;
 			if (dep->tag == TAG_REMOTE)
 				mnt->tag = TAG_REMOTE;
 		}
@@ -1274,17 +1326,35 @@ tag_mount (Mount *mnt,
 		return;
 	}
 
+	if ((tag == TAG_VIRTUAL)
+	    && (mnt->tag == TAG_TIMEOUT)) {
+		nih_debug ("%s is not virtual, inherited local (with timeout)",
+			   MOUNT_NAME (mnt));
+		return;
+	}
+
 	if ((tag == TAG_LOCAL)
 	    && (mnt->tag == TAG_REMOTE)) {
 		nih_debug ("%s is not local, inherited remote",
 			   MOUNT_NAME (mnt));
 	}
 
+	if ((tag == TAG_TIMEOUT)
+	    && (mnt->tag == TAG_REMOTE)) {
+		nih_debug ("%s is not local (with timeout), inherited remote",
+			   MOUNT_NAME (mnt));
+	}
+
 	/* Set the tag, then if it's one we'd normally inherit, set it on
 	 * any mount point that depends on this one.
 	 */
 	mnt->tag = tag;
 
+	/* TAG_TIMEOUT is TAG_LOCAL with a timeout. timeout cannot be
+	 * inherited but local could be */
+	if (tag == TAG_TIMEOUT)
+		tag = TAG_LOCAL;
+
 	if ((tag == TAG_LOCAL)
 	    || (tag == TAG_REMOTE)) {
 		NIH_LIST_FOREACH (mounts, iter) {
@@ -1384,6 +1454,19 @@ mounted (Mount *mnt)
 	case TAG_LOCAL:
 		num_local_mounted++;
 		break;
+	case TAG_TIMEOUT:
+		num_local_mounted++;
+		num_timeout_mounted++;
+		if (num_timeout_mounted == num_timeout) {
+			nih_message (_("\n %s finished! "), "local_timeout");
+			/* Stop the timeout waiting for device to get ready"
+			 */
+			if (device_ready_timer) {
+				nih_free (device_ready_timer);
+				device_ready_timer = NULL;
+			}
+		}
+		break;
 	case TAG_REMOTE:
 		num_remote_mounted++;
 		break;
@@ -1420,6 +1503,10 @@ skip_mount (Mount *mnt)
 	case TAG_LOCAL:
 		num_local--;
 		break;
+	case TAG_TIMEOUT:
+		num_local--;
+		num_timeout--;
+		break;
 	case TAG_REMOTE:
 		num_remote--;
 		break;
@@ -1519,6 +1606,85 @@ trigger_events (void)
 		   num_swap_mounted, num_swap);
 }
 
+/* 
+ * is_device_ready:
+ *
+ * @data: Not used
+ * @timer: timer which was preconfigured and on the expiry of which this
+ * function is called.
+ *
+ * This function is called to check if any device marked with the "timeout"
+ * option in the /etc/fstab is not ready when the preconfigured @timer
+ * expires.
+ * It emits a "device-not-ready" event when it finds that such a device is not
+ * yet ready.
+ */
+void
+is_device_ready (void * data,
+		 NihTimer *timer)
+{
+	device_ready_timer = NULL;
+	NIH_LIST_FOREACH (mounts, iter) {
+		Mount *mnt = (Mount *)iter;
+
+		if (mnt->mounted)
+			continue;
+		if (mnt->tag != TAG_TIMEOUT)
+			continue;
+		/* If udev was not able to identify the expected device within
+		 * ROOTDELAY time, something is wrong */
+		if ((! mnt->ready)
+		    && (! mnt->nodev)
+		    && (! is_remote (mnt))
+		    && ((! strncmp (mnt->device, "/dev/", 5))
+			|| (! strncmp (mnt->device, "UUID=", 5))
+			|| (! strncmp (mnt->device, "LABEL=", 6))))
+		{
+			nih_message ("%s device not ready in ROOTDELAY sec", MOUNT_NAME (mnt));
+			emit_event ("device-not-ready", mnt);
+		}
+	}
+}
+
+/* 
+ * activate_timer:
+ *
+ * This function is called to start a timer when the first device with
+ * "timeout" option is found.
+ */
+void 
+activate_timer (void)
+{
+	NIH_LIST_FOREACH (mounts, iter) {
+		Mount *mnt = (Mount *)iter;
+
+		if (mnt->mounted)
+			continue;
+		if (mnt->tag != TAG_TIMEOUT)
+			continue;
+		/* If there's an underlying device that udev is going to deal with,
+		 * and it's not a virtual or remote filesystem, we wait for the udev
+		 * watcher to mark it ready within ROOTDELAY time.
+		 */
+		if ((! mnt->ready)
+		    && (! mnt->nodev)
+		    && (! is_remote (mnt))
+		    && ((! strncmp (mnt->device, "/dev/", 5))
+			|| (! strncmp (mnt->device, "UUID=", 5))
+			|| (! strncmp (mnt->device, "LABEL=", 6))))
+		{
+			if (!dev_wait_time)
+				dev_wait_time = ROOTDELAY;
+			nih_debug ("Shall wait for device: %s for %d seconds, starting timer", MOUNT_NAME (mnt), dev_wait_time);
+			device_ready_timer = NIH_MUST (nih_timer_add_timeout (NULL, 
+						dev_wait_time, is_device_ready, NULL));
+			break;
+
+		}
+
+	}
+
+}
 
 void
 try_mounts (void)
@@ -1542,8 +1708,13 @@ try_mounts (void)
 			}
 		}
 
-		if (all)
+		if (all) {
+			if (control_server) {
+				dbus_server_disconnect (control_server);
+				dbus_server_unref (control_server);
+			}
 			nih_main_loop_exit (EXIT_OK);
+		}
 	}
 }
 
@@ -1831,7 +2002,7 @@ run_mount (Mount *mnt,
 		 */
 		if ((errno != EEXIST)
 		    && has_option (mnt, "optional", FALSE)) {
-			nih_debug ("%s: mountpoint doesn't exist, ignoring",
+			nih_debug ("%s: mountpoint doesn't exist; ignoring",
 				   MOUNT_NAME (mnt));
 			mounted (mnt);
 			return;
@@ -1839,7 +2010,7 @@ run_mount (Mount *mnt,
 	}
 
 	opts = cut_options (NULL, mnt, "showthrough", "optional",
-			    "bootwait", "nobootwait",
+			    "bootwait", "nobootwait", "timeout",
 			    NULL);
 	if (mnt->mounted && (! fake)) {
 		char *tmp;
@@ -3018,37 +3189,37 @@ plymouth_update (int only_clear)
 	/* Display that message */
 	switch (plymouth_error) {
 	case ERROR_BORED:
-		message = NIH_MUST (nih_sprintf (NULL,_("The disk drive for %s is not ready yet or not present"),
+		message = NIH_MUST (nih_sprintf (NULL,_("The disk drive for %s is not ready yet or not present."),
 						 MOUNT_NAME (plymouth_mnt)));
 		keys_message = NIH_MUST (nih_sprintf (NULL, "keys:%s",
-						      _("Continue to wait; or Press S to skip mounting or M for manual recovery")));
+						      _("Continue to wait, or Press S to skip mounting or M for manual recovery")));
 		plymouth_keys = "SsMm";
 		break;
 	case ERROR_FSCK_FAILED:
-		message = NIH_MUST (nih_sprintf (NULL,_("Errors were found while checking the disk drive for %s"),
+		message = NIH_MUST (nih_sprintf (NULL,_("Errors were found while checking the disk drive for %s."),
 						 MOUNT_NAME (plymouth_mnt)));
 		keys_message = NIH_MUST (nih_sprintf (NULL, "keys:%s",
-						      _("Press F to attempt to fix the errors, I to ignore, S to skip mounting or M for manual recovery")));
+						      _("Press F to attempt to fix the errors, I to ignore, S to skip mounting, or M for manual recovery")));
 		plymouth_keys = "FfIiSsMm";
 		break;
 	case ERROR_FSCK_FAILED_HARD:
-		message = NIH_MUST (nih_sprintf (NULL,_("Serious errors were found while checking the disk drive for %s"),
+		message = NIH_MUST (nih_sprintf (NULL,_("Serious errors were found while checking the disk drive for %s."),
 						 MOUNT_NAME (plymouth_mnt)));
 		keys_message = NIH_MUST (nih_sprintf (NULL, "keys:%s",
-						      _("Press I to ignore, S to skip mounting or M for manual recovery")));
+						      _("Press I to ignore, S to skip mounting, or M for manual recovery")));
 		plymouth_keys = "IiSsMm";
 		break;
 	case ERROR_MOUNT_FAILED:
-		message = NIH_MUST (nih_sprintf (NULL, _("An error occurred while mounting %s"),
+		message = NIH_MUST (nih_sprintf (NULL, _("An error occurred while mounting %s."),
 						 MOUNT_NAME (plymouth_mnt)));
 		keys_message = NIH_MUST (nih_sprintf (NULL, "keys:%s",
 						      _("Press S to skip mounting or M for manual recovery")));
 		plymouth_keys = "SsMm";
 		break;
 	case ERROR_FSCK_IN_PROGRESS:
-		message = NIH_MUST (nih_strdup (NULL, _("Your disk drives are being checked for errors, this may take some time")));
+		message = NIH_MUST (nih_strdup (NULL, _("Checking disk drives for errors. This may take several minutes.")));
 		keys_message = NIH_MUST (nih_sprintf (NULL, "keys:%s",
-						      _("Press C to cancel all checks currently in progress")));
+						      _("Press C to cancel all checks in progress")));
 		plymouth_keys = "Cc";
 		break;
 	default:
@@ -3194,8 +3365,230 @@ plymouth_answer (void *             user_data,
 	}
 }
 
+/*
+ * set_dev_wait_time:
+ *
+ * @option: Not used
+ * @arg: String receieved at the mountall command line by nih_option_parser
+ * and passed to this function for parsing.
+ *
+ * This function is used to set the "dev_wait_time option" argument specified
+ * at the mountall command line. It is called by the nih_option_parser for
+ * parsing the argument string specified by @arg
+ *
+ */
+int
+set_dev_wait_time (NihOption *option,
+		   const char *arg)
+{
+        char * end_ptr;
+        dev_wait_time = strtol (arg, &end_ptr, 10);
+        int err = 0;
+        if (dev_wait_time <= 0) {
+                nih_error (_("\n Legal values of dev-wait-time lie between 1sec to 2147483647 sec"));
+                err = -1;
+        }
+        else if ((dev_wait_time == LONG_MIN) || (dev_wait_time == LONG_MAX)) {
+                nih_error (_("\n Legal values of dev-wait-time lie between 1sec to 2147483647 sec"));
+                err = -1;
+        }
+        else if (*end_ptr != '\0') {
+                nih_error (_("\n Legal values of dev-wait-time lie between 1sec to 2147483647 sec"));
+                err = -1;
+        }
+        return err;
+}
 
 
+/**
+ * stop_mountpoint_timer:
+ * @mountpoint: mountpoint whose timer you want to stop.
+ *
+ * This function is called to stop a previously started timer of a device
+ * Note that for this function to be successful, @mountpoint should be a
+ * mountpoint corresponding to a device whose "timeout" option is set in the
+ * /etc/fstab.
+ *
+ * Returns 0 on stopping the timer and -1 on not doing so.
+ *
+ **/
+int
+stop_dev_timer (const char *mountpoint)
+{
+	int ret = -1;
+
+	nih_assert (mountpoint != NULL);
+
+	if (strlen (mountpoint) == 0) {
+		nih_message (_("Empty mountpoint specified"));
+		return ret;
+	}
+
+	NIH_LIST_FOREACH (mounts, iter) {
+		Mount *mnt = (Mount *)iter;
+
+		if (mnt->mounted)
+			continue;
+		if (mnt->tag != TAG_TIMEOUT)
+			continue;
+
+                if ((! mnt->ready)
+                    && (! mnt->nodev)
+                    && (! is_remote (mnt))
+                    && ((! strncmp (mnt->device, "/dev/", 5))
+                        || (! strncmp (mnt->device, "UUID=", 5))
+                        || (! strncmp (mnt->device, "LABEL=", 6)))
+                    && (! strcmp (mnt->mountpoint, mountpoint)))
+                {
+			if (device_ready_timer) {
+				nih_debug ("Stopping the timer for the device:"
+					"%s for %d seconds, starting timer", 
+					MOUNT_NAME (mnt), dev_wait_time);
+	                        nih_free (device_ready_timer);
+				device_ready_timer = NULL;
+				ret = 0;
+			}
+                        break;
+
+                }
+	}
+	return ret;
+}
+
+/**
+ * restart_dev_timer:
+ ** @mountpoint: mountpoint whose timer you want to stop.
+ *
+ * This function is called to restart a previously stopped timer of a device
+ * Note that for this function to be successful, @mountpoint should be a
+ * mountpoint corresponding to a device whose "timeout" option is set in the
+ * /etc/fstab.
+ *
+ * Returns 0 on restarting the timer and -1 on not doing so.
+ *
+ **/
+int
+restart_dev_timer (const char *mountpoint)
+{
+	int ret = -1;
+
+	nih_assert (mountpoint != NULL);
+
+	if (strlen (mountpoint) == 0) {
+		nih_message (_("Empty mountpoint specified"));
+		return ret;
+	}
+
+	NIH_LIST_FOREACH (mounts, iter) {
+		Mount *mnt = (Mount *)iter;
+
+		if (mnt->mounted)
+			continue;
+		if (mnt->tag != TAG_TIMEOUT)
+			continue;
+
+                if ((! mnt->ready)
+                    && (! mnt->nodev)
+                    && (! is_remote (mnt))
+                    && ((! strncmp (mnt->device, "/dev/", 5))
+                        || (! strncmp (mnt->device, "UUID=", 5))
+                        || (! strncmp (mnt->device, "LABEL=", 6)))
+                    && (! strcmp (mnt->mountpoint, mountpoint)))
+                {
+
+			if (!dev_wait_time)
+				dev_wait_time = ROOTDELAY;
+			if (!device_ready_timer) {
+				nih_debug ("Shall wait for device: %s for %d "
+					"seconds, starting timer", 
+					MOUNT_NAME (mnt), dev_wait_time);
+				device_ready_timer = NIH_MUST (nih_timer_add_timeout (NULL, 
+					dev_wait_time, is_device_ready, NULL));
+				ret = 0;
+			} 
+			break;
+
+                }
+	}
+	return ret;
+}
+
+/**
+ * change_mount_device:
+ * @devname: name of the new device.
+ * @path: full path to a previously existing mountpoint.
+ *
+ *  This function is called to change the device that would be mounted at a
+ *  given mountpoint without calling this function. Note that @path should be
+ *  what mountall has read from fstab or what mountall already knows for
+ *  mounting a device.
+ *
+ *  Return 0 on successfully changing the device and -1 on not doing so.
+ **/
+int
+change_mount_device (const char *devname,
+	             const char *path)
+{
+	int ret = -1;
+
+	nih_assert (devname != NULL);
+	nih_assert (path != NULL);
+
+	if (strlen (devname) == 0) {
+		nih_error(_("Empty device name specified"));
+		return ret;
+	}
+	if (strlen (path) == 0) {
+		nih_error (_("Empty mount point specified"));
+		return ret;
+	}
+
+	NIH_LIST_FOREACH (mounts, iter) {
+		Mount *mnt = (Mount *)iter;
+
+		if (mnt->mounted)
+			continue;
+		if (mnt->tag != TAG_TIMEOUT)
+			continue;
+
+                if ((! mnt->ready)
+                    && (!mnt->nodev)
+                    && (!is_remote (mnt))
+                    && ((!strncmp (mnt->device, "/dev/", 5))
+                        || (!strncmp (mnt->device, "UUID=", 5))
+                        || (!strncmp (mnt->device, "LABEL=", 6)))
+                    )
+                {
+			/* Change the device to be mounted on a pre-registered
+			 * mountpoint
+			 */
+			if ((!strcmp (mnt->mountpoint, path))) {
+				/* Change only if the requested device is
+				 * really any different than whats stored
+				 */
+				if (!strcmp (mnt->device, devname)) {
+					char * newdev;
+					ret = 0;
+					newdev  = nih_strdup (mounts, devname);
+					if (!newdev)  {
+						nih_error_raise_no_memory ();
+						nih_error (("Could not change the mountpoint "
+								"for device: %s"), devname);
+						ret = -1;
+					}
+					else {
+						nih_free (mnt->device);
+						mnt->device = newdev;
+						emit_event ("changed-device", mnt);
+					}
+				}
+				break;
+			}
+                }
+	}
+	return ret;
+}
+
 /**
  * options:
  *
@@ -3210,7 +3603,8 @@ static NihOption options[] = {
 	  NULL, NULL, &fsck_fix, NULL },
 	{ 0, "no-events", N_("Do not emit events after mounting filesystems"),
 	  NULL, NULL, &no_events, NULL },
-
+	{ 0, "dev-wait-time", N_("In case of (bootwait,timeout): specify the time to wait for device to be detected"),
+          NULL, N_("value in seconds (Default is 30 seconds, Legal value between 1second to 2147483647 seconds)"), &dev_wait_time, set_dev_wait_time},
 	NIH_OPTION_LAST
 };
 
@@ -3313,7 +3707,7 @@ main (int   argc,
 
 	/* Sanity check, the root filesystem should be already mounted */
 	root = find_mount ("/");
-	if (! root->mounted) {
+	if (! root || !root->mounted ) {
 		nih_fatal ("%s", _("root filesystem isn't mounted"));
 		exit (EXIT_ERROR);
 	}
@@ -3394,6 +3788,25 @@ main (int   argc,
 
 	/* See what's already mounted */
 	mark_mounted ();
+	/* Create a listening server for private connections. */
+	while ((ret = control_server_open ()) < 0) {
+		NihError *err;
+
+		err = nih_error_get ();
+		if (err->number != ENOMEM) {
+			nih_warn ("%s: %s", _("Unable to listen for private"
+						"connections"), err->message);
+			nih_free (err);
+			break;
+		}
+		nih_free (err);
+	}
+	if (!ret)
+		emit_event("mountallServer", NULL);
+	/* Activate the timer for a fs that is local, unmounted and waits for
+	 * a device to be ready, before it can be mounted onto it. Timer on
+	 * only for fs not marked with a "nobootwait=1" */
+	activate_timer();
 
 	/* See what we can mount straight away, and then schedule the same
 	 * function to be run each time through the main loop.
